/*
 * acompare.cpp
 *
 *  Created on: 24. 2. 2014
 *      Author: Jan Travnicek
 */

#include <version.hpp>

#include <tclap/CmdLine.h>
#include <global/GlobalData.h>
#include <alib/measure>

#include <exception/CommonException.h>
#include <lexer/Lexer.h>
#include <parser/Parser.h>

#include <factory/XmlDataFactory.hpp>
#include <abstraction/ValueProvider.hpp>

int main ( int argc, char * * argv ) {
	try {
		common::GlobalData::argc = argc;
		common::GlobalData::argv = argv;

		TCLAP::CmdLine cmd ( "Regexp derivative compute binary", ' ', ALIB_VERSION_INFO );

		TCLAP::UnlabeledValueArg < std::string > input1 ( "input1", "The first input to compare", true, "-", "file" );
		cmd.add ( input1 );

		TCLAP::UnlabeledValueArg < std::string > input2 ( "input2", "The second input to compare", true, "-", "file" );
		cmd.add ( input2 );

		TCLAP::SwitchArg measure ( "m", "measure", "Measure times", false );
		cmd.add ( measure );

		TCLAP::SwitchArg verbose ( "v", "verbose", "Be verbose", false );
		cmd.add ( verbose );

		cmd.parse ( argc, argv );

		if ( verbose.isSet ( ) )
			common::GlobalData::verbose = true;

		if ( measure.isSet ( ) )
			common::GlobalData::measure = true;

		cli::Environment environment;
		environment.setBinding ( "stdin1", input1.getValue ( ) );
		environment.setBinding ( "stdin2", input2.getValue ( ) );
		environment.setBinding ( "stdout", "-" );

		measurements::start ( "Overal", measurements::Type::OVERALL );
		measurements::start ( "Input read", measurements::Type::AUXILIARY );

		cli::Parser parser ( cli::Lexer ( "execute <#stdin1 > $input1" ) );
		parser.parse ( )->run ( environment );

		parser = cli::Parser ( cli::Lexer ( "execute <#stdin2 > $input2" ) );
		parser.parse ( )->run ( environment );

		std::string input1Type = environment.getVariable ( "input1" )->getReturnType ( );
		std::string input2Type = environment.getVariable ( "input2" )->getReturnType ( );
		bool isAutomaton = input1Type.find ( "automaton::" ) == 0 && input2Type.find ( "automaton::" ) == 0;
		bool isGrammar = input1Type.find ( "grammar::" ) == 0 && input2Type.find ( "grammar::" ) == 0;
		bool isString = input1Type.find ( "string::" ) == 0 && input2Type.find ( "string::" ) == 0;

		measurements::end ( );
		measurements::start ( "Compare", measurements::Type::MAIN );

		if ( isAutomaton ) {
			parser = cli::Parser ( cli::Lexer ( "execute compare::AutomatonDiff $input1 $input2 > $output" ) );
		} else if ( isGrammar ) {
			parser = cli::Parser ( cli::Lexer ( "execute compare::GrammarDiff $input1 $input2 > $output" ) );
		} else if ( isString ) {
			parser = cli::Parser ( cli::Lexer ( "execute comapre::StringDiff $input1 $input2 > $output" ) );
		} else {
			throw exception::CommonException ( "Only automata, grammars and strings can be compared." );
		}
		parser.parse ( )->run ( environment );

		std::string result = std::dynamic_pointer_cast < abstraction::ValueProvider < std::string > > ( environment.getVariable ( "output" ) )->getValue ( true );

		if ( result != "" )
			common::Streams::out << result << std::endl;

		measurements::end ( );
		measurements::end ( );

		return result != "";
	} catch ( const exception::CommonException & exception ) {
		factory::XmlDataFactory::toStdout ( exception );
		return 1;
	} catch ( const TCLAP::ArgException & exception ) {
		common::Streams::err << exception.error ( ) << std::endl;
		return 2;
	} catch ( const std::exception & exception ) {
		common::Streams::err << "Exception caught: " << exception.what ( ) << std::endl;
		return 3;
	} catch ( ... ) {
		common::Streams::err << "Unknown exception caught." << std::endl;
		return 127;
	}
}
