#include <stringology/matching/LevenshteinMatchingAutomaton.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <string/LinearString.h>
#include <automaton/simplify/UnreachableStatesRemover.h>

#include "LevenshteinMatchingAutomatonTest.h"


CPPUNIT_TEST_SUITE_NAMED_REGISTRATION ( LevenshteinMatchingAutomatonTest, "stringology" );
CPPUNIT_TEST_SUITE_REGISTRATION ( LevenshteinMatchingAutomatonTest );


void LevenshteinMatchingAutomatonTest::testSimpleConstruction() {
    ext::set<char> alphabet{'a', 'b', 'c', 'd'};
    string::LinearString <char> input_string(alphabet, ext::vector<char>{'a', 'b', 'c'});
    auto resulting_automata = stringology::matching::LevenshteinMatchingAutomaton::construct(input_string, 2);

    typedef ext::pair<unsigned int, unsigned int> State;

    automaton::EpsilonNFA < char, void, State > res(ext::make_pair(0,0));
    res.setInputAlphabet(ext::set<char>{'a', 'b', 'c', 'd'});

    State q0 = ext::make_pair(0,0);
    State q1 = ext::make_pair(1,0);
    State q2 = ext::make_pair(2,0);
    State q3 = ext::make_pair(3,0);
    State q4 = ext::make_pair(1,1);
    State q5 = ext::make_pair(2,1);
    State q6 = ext::make_pair(3,1);
    State q7 = ext::make_pair(2,2);
    State q8 = ext::make_pair(3,2);

    res.setStates(ext::set<State> {q0, q1, q2, q3, q4, q5, q6, q7, q8});
    res.setFinalStates(ext::set<State> {q3, q6, q8});

    res.addTransition(q0, 'a', q1); // vertical transitions (exact matching automata)

    res.addTransition(q1, 'b', q2);
    res.addTransition(q4, 'b', q5);

    res.addTransition(q2, 'c', q3);
    res.addTransition(q5, 'c', q6);
    res.addTransition(q7, 'c', q8);

    res.addTransition(q0, 'a', q0); // loops in initial state
    res.addTransition(q0, 'b', q0);
    res.addTransition(q0, 'c', q0);
    res.addTransition(q0, 'd', q0);

    res.addTransition(q0, 'b', q4); // diagonal transitions representing replace
    res.addTransition(q0, 'c', q4);
    res.addTransition(q0, 'd', q4);

    res.addTransition(q0, q4); // deletion

    res.addTransition(q1, 'a', q5);
    res.addTransition(q1, 'c', q5);
    res.addTransition(q1, 'd', q5);
    res.addTransition(q4, 'a', q7);
    res.addTransition(q4, 'c', q7);
    res.addTransition(q4, 'd', q7);

    res.addTransition(q1, q5); // deletion
    res.addTransition(q4, q7);

    res.addTransition(q2, 'a', q6);
    res.addTransition(q2, 'b', q6);
    res.addTransition(q2, 'd', q6);
    res.addTransition(q5, 'a', q8);
    res.addTransition(q5, 'b', q8);
    res.addTransition(q5, 'd', q8);

    res.addTransition(q2, q6); // deletion
    res.addTransition(q5, q8);

    res.addTransition(q1, 'a', q4); // insertions
    res.addTransition(q1, 'b', q4);
    res.addTransition(q1, 'c', q4);
    res.addTransition(q1, 'd', q4);

    res.addTransition(q2, 'a', q5);
    res.addTransition(q2, 'b', q5);
    res.addTransition(q2, 'c', q5);
    res.addTransition(q2, 'd', q5);

    res.addTransition(q5, 'a', q7);
    res.addTransition(q5, 'b', q7);
    res.addTransition(q5, 'c', q7);
    res.addTransition(q5, 'd', q7);

    CPPUNIT_ASSERT(res == automaton::simplify::UnreachableStatesRemover::remove(resulting_automata));
}

void LevenshteinMatchingAutomatonTest::testSimpleWildcardConstruction() {
  ext::set<char> alphabet{'a', 'b', '@'};
  string::WildcardLinearString <char> input_string(alphabet, ext::vector<char>{'a', '@', 'b'}, '@');
  auto resulting_automata = stringology::matching::LevenshteinMatchingAutomaton::construct(input_string, 2);

  typedef ext::pair<unsigned int, unsigned int> State;

  automaton::EpsilonNFA < char, void, State > res(ext::make_pair(0,0));
  res.setInputAlphabet(alphabet);

  State q0 = ext::make_pair(0,0);
  State q1 = ext::make_pair(1,0);
  State q2 = ext::make_pair(2,0);
  State q3 = ext::make_pair(3,0);
  State q4 = ext::make_pair(1,1);
  State q5 = ext::make_pair(2,1);
  State q6 = ext::make_pair(3,1);
  State q7 = ext::make_pair(2,2);
  State q8 = ext::make_pair(3,2);

  res.setStates(ext::set<State> {q0, q1, q2, q3, q4, q5, q6, q7, q8});
  res.setFinalStates(ext::set<State> {q3, q6, q8});

  res.addTransition(q0, 'a', q0); // initial loops
  res.addTransition(q0, 'b', q0);

  res.addTransition(q0, 'a', q1); // 3 simple matching automatas

  res.addTransition(q1, 'a', q2);
  res.addTransition(q1, 'b', q2);
  res.addTransition(q4, 'a', q5);
  res.addTransition(q4, 'b', q5);

  res.addTransition(q2, 'b', q3);
  res.addTransition(q5, 'b', q6);
  res.addTransition(q7, 'b', q8);

  res.addTransition(q0, 'b', q4); // error transitions for replace

  res.addTransition(q2, 'a', q6);
  res.addTransition(q5, 'a', q8);

  res.addTransition(q0, q4); // delete transition
  res.addTransition(q1, q5);
  res.addTransition(q2, q6);

  res.addTransition(q4, q7);
  res.addTransition(q5, q8);

  res.addTransition(q1, 'a', q4);
  res.addTransition(q1, 'b', q4);

  res.addTransition(q2, 'a', q5);
  res.addTransition(q2, 'b', q5);

  res.addTransition(q5, 'a', q7);
  res.addTransition(q5, 'b', q7);

  CPPUNIT_ASSERT(res == automaton::simplify::UnreachableStatesRemover::remove(resulting_automata));
}

void LevenshteinMatchingAutomatonTest::setUp() { }

void LevenshteinMatchingAutomatonTest::tearDown() { }
