#include "ModelBoxTest.h"

#include <Models/InputModelBox.hpp>
#include <Models/OutputModelBox.hpp>
#include <Models/AlgorithmModelBox.hpp>
#include <Algorithm/Registry.hpp>
#include <Utils.hpp>

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION ( ModelBoxTest, "model" );
CPPUNIT_TEST_SUITE_REGISTRATION ( ModelBoxTest );

void ModelBoxTest::setUp ( ) {
	Registry::initialize();
}

void ModelBoxTest::tearDown ( ) {
	Registry::deinitialize();
}

void ModelBoxTest::modelBoxBasicProperties ( ) {
	auto input = std::make_unique<InputModelBox>();
	auto output = std::make_unique<OutputModelBox>();
	auto algo = std::make_unique<AlgorithmModelBox>(Registry::getAlgorithm("automaton::determinize::Determinize"));

	CPPUNIT_ASSERT(input->getMaxInputCount() == 0);
	CPPUNIT_ASSERT(input->canHaveOutput());

	CPPUNIT_ASSERT(output->getMaxInputCount() == 1);
	CPPUNIT_ASSERT(output->canHaveOutput() == false);

	CPPUNIT_ASSERT(algo->getMaxInputCount() == 1);
	CPPUNIT_ASSERT(algo->canHaveOutput());

	CPPUNIT_ASSERT(input->getType() == ModelType::Input);
	CPPUNIT_ASSERT(input->getName() == "INPUT");

	CPPUNIT_ASSERT(output->getType() == ModelType::Output);
	CPPUNIT_ASSERT(output->getName() == "OUTPUT");

	CPPUNIT_ASSERT(algo->getType() == ModelType::Algorithm);
	CPPUNIT_ASSERT(algo->getName() == "Determinize");

	CPPUNIT_ASSERT(algo->getAlgorithm() == Registry::getAlgorithm("automaton::determinize::Determinize"));
}

void ModelBoxTest::modelBoxConnectDisconnectBasic ( ) {
	auto input = std::make_unique<InputModelBox>();
	auto output = std::make_unique<OutputModelBox>();

	CPPUNIT_ASSERT(output->evaluate() == nullptr);

	input->setAutomaton(Utils::generateRandomAutomaton());

	CPPUNIT_ASSERT(output->evaluate() == nullptr);

	ModelBox::connect(input.get(), output.get(), 0);

	CPPUNIT_ASSERT(output->evaluate() == input->getAutomaton());

	ModelBox::disconnect(input.get(), output.get(), 0);

	CPPUNIT_ASSERT(output->evaluate() == nullptr);
}

void ModelBoxTest::modelBoxConnectDisconnectMultipleFrom ( ) {
	auto input = std::make_unique<InputModelBox>();
	auto output1 = std::make_unique<OutputModelBox>();
	auto output2 = std::make_unique<OutputModelBox>();

	CPPUNIT_ASSERT(output1->evaluate() == nullptr);
	CPPUNIT_ASSERT(output2->evaluate() == nullptr);

	input->setAutomaton(Utils::generateRandomAutomaton());

	CPPUNIT_ASSERT(output1->evaluate() == nullptr);
	CPPUNIT_ASSERT(output2->evaluate() == nullptr);

	ModelBox::connect(input.get(), output1.get(), 0);

	CPPUNIT_ASSERT(output1->evaluate() == input->getAutomaton());
	CPPUNIT_ASSERT(output2->evaluate() == nullptr);

	ModelBox::connect(input.get(), output2.get(), 0);

	CPPUNIT_ASSERT(output1->evaluate() == input->getAutomaton());
	CPPUNIT_ASSERT(output2->evaluate() == input->getAutomaton());

	ModelBox::disconnect(input.get(), output1.get(), 0);

	CPPUNIT_ASSERT(output1->evaluate() == nullptr);
	CPPUNIT_ASSERT(output2->evaluate() == input->getAutomaton());
}

void ModelBoxTest::modelBoxConnectDisconnectMultipleTo ( ) {
	auto input1 = std::make_unique<InputModelBox>();
	auto input2 = std::make_unique<InputModelBox>();
	auto output = std::make_unique<OutputModelBox>();

	CPPUNIT_ASSERT(output->evaluate() == nullptr);

	ModelBox::connect(input1.get(), output.get(), 0);
	CPPUNIT_ASSERT_THROW(ModelBox::connect(input2.get(), output.get(), 0), std::runtime_error);
}

void ModelBoxTest::modelBoxConnectDisconnectMultipleExisting ( ) {
	auto input1 = std::make_unique<InputModelBox>();
	auto input2 = std::make_unique<InputModelBox>();
	auto output = std::make_unique<OutputModelBox>();

	ModelBox::connect(input1.get(), output.get(), 0);
	ModelBox::disconnect(input1.get(), output.get(), 0);
	CPPUNIT_ASSERT_NO_THROW(ModelBox::connect(input2.get(), output.get(), 0));
}

void ModelBoxTest::modelBoxConnectDisconnectNonexistentSlot ( ) {
	auto input = std::make_unique<InputModelBox>();
	auto output = std::make_unique<OutputModelBox>();

	CPPUNIT_ASSERT_THROW(ModelBox::connect(input.get(), output.get(), 1), std::runtime_error);
}

void ModelBoxTest::modelBoxConnectDisconnectNonexistentConnection ( ) {
	auto input = std::make_unique<InputModelBox>();
	auto output = std::make_unique<OutputModelBox>();

	CPPUNIT_ASSERT_THROW(ModelBox::disconnect(input.get(), output.get(), 0), std::runtime_error);
}
