/*
 * atrim.cpp
 *
 *  Created on: 26. 3. 2014
 *	  Author: Tomas Pecka
 */

#include <version.hpp>

#include <tclap/CmdLine.h>
#include <global/GlobalData.h>
#include <alib/measure>

#include <exception/CommonException.h>
#include <lexer/Lexer.h>
#include <parser/Parser.h>

#include <factory/XmlDataFactory.hpp>

std::string trimGrammar ( bool del_unreachable, bool del_unproductive ) {
	if( del_unreachable && del_unproductive )
		return "execute grammar::simplify::Trim $input > $output";
	if( del_unreachable )
		return "execute grammar::simplify::UnreachableSymbolsRemover $input > $output";
	if( del_unproductive )
		return "execute grammar::simplify::UnproductiveSymbolsRemover $input > $output";
	return "execute grammar::simplify::Trim $input > $output";
}

std::string trimAutomaton ( bool del_unreachable, bool del_useless ) {
	if( del_unreachable && del_useless )
		return "execute automaton::simplify::Trim $input > $output";
	if( del_unreachable )
		return "execute automaton::simplify::UnreachableStatesRemover $input > $output";
	if( del_useless )
		return "execute automaton::simplify::UselessStatesRemover $input > $output";
	return "execute automaton::simplify::Trim $input > $output";
}

std::string efficientTrimAutomaton ( bool del_unreachable, bool del_useless ) {
	if( del_unreachable && del_useless )
		return "execute automaton::simplify::efficient::Trim $input > $output";
	if( del_unreachable )
		return "execute automaton::simplify::efficient::UnreachableStatesRemover $input > $output";
	if( del_useless )
		return "execute automaton::simplify::efficient::UselessStatesRemover $input > $output";
	return "execute automaton::simplify::efficient::Trim $input > $output";
}

std::string optimizeRegExp ( ) {
	return "execute regexp::simplify::RegExpOptimize $input > $output";
}

int main(int argc, char* argv[]) {
	try {
		common::GlobalData::argc = argc;
		common::GlobalData::argv = argv;

		TCLAP::CmdLine cmd("Removes unreachable and useless states from FSM, productive and unreachable nonterminals from CFG. Simplifies representation of RE", ' ', ALIB_VERSION_INFO);

		TCLAP::SwitchArg useless(		"u",	"useless",	"Removes useless states (or symbols). (works with FSM, FTA or CFG)"	);
		cmd.add( useless );

		TCLAP::SwitchArg unreachable(		"r",	"unreachable",	"Removes unreachable states (or symbols). (works with FSM, FTA or CFG)"	);
		cmd.add( unreachable );

		TCLAP::SwitchArg simplify(		"s",	"simplify",	"Simplifies representation. (works with RE only)"			);
		cmd.add( simplify );

		TCLAP::ValueArg<std::string> input(	"i",	"input",	"Input to trim",							false,	"-",	"file");
		cmd.add( input );

		TCLAP::SwitchArg efficient(		"e",	"efficient",	"Use efficient implementation",		false);
		cmd.add( efficient );

		TCLAP::SwitchArg measure(		"m",	"measure",	"Measure times",		false);
		cmd.add( measure );

		TCLAP::SwitchArg verbose(		"v",	"verbose",	"Be verbose",			false);
		cmd.add( verbose );

		cmd.parse(argc, argv);

		if(verbose.isSet())
			common::GlobalData::verbose = true;
		if(measure.isSet())
			common::GlobalData::measure = true;

		cli::Environment environment;
		environment.setBinding ( "stdin", input.getValue ( ) );
		environment.setBinding ( "stdout", "-" );

		measurements::start("Overal", measurements::Type::OVERALL);
		measurements::start("Input read", measurements::Type::AUXILIARY);

		cli::Parser parser ( cli::Lexer ( "execute <#stdin > $input" ) );
		parser.parse ( )->run ( environment );

//		parser = cli::Parser ( cli::Lexer ( "export type = typename ( $input )" ) );
		std::string inputType = environment.getVariable ( "input" )->getReturnType ( );
		bool isAutomaton = inputType.find ( "automaton::" ) == 0;
		bool isGrammar = inputType.find ( "grammar::" ) == 0;
		bool isRegExp = inputType.find ( "regexp::" ) == 0;

		measurements::end();
		measurements::start("Algorithm", measurements::Type::MAIN);

		if( isAutomaton ) {
			if(efficient.getValue()) {
				parser = cli::Parser ( cli::Lexer ( efficientTrimAutomaton ( unreachable.getValue(), useless.getValue() ) ) );
			} else {
				parser = cli::Parser ( cli::Lexer ( trimAutomaton ( unreachable.getValue(), useless.getValue() ) ) );
			}
		} else if( isGrammar ) {
			parser = cli::Parser ( cli::Lexer ( trimGrammar ( unreachable.getValue(), useless.getValue() ) ) );
		} else if( isRegExp ) {
			parser = cli::Parser ( cli::Lexer ( optimizeRegExp ( ) ) );
		} else {
			throw exception::CommonException( "Input not recognized." );
		}
		parser.parse ( )->run ( environment );

		measurements::end();
		measurements::start("Output write", measurements::Type::AUXILIARY);

		parser = cli::Parser ( cli::Lexer ( "execute $output >#stdout" ) );
		parser.parse ( )->run ( environment );

		measurements::end();
		measurements::end();

		if(measure.getValue()) common::Streams::measure << measurements::results() << std::endl;

		return 0;
	} catch( const exception::CommonException & exception ) {
		factory::XmlDataFactory::toStdout( exception );
		return 1;
	} catch(const TCLAP::ArgException& exception) {
		common::Streams::out << exception.error() << std::endl;
		return 2;
	} catch (const std::exception& exception) {
		common::Streams::err << "Exception caught: " << exception.what() << std::endl;
		return 3;
	} catch(...) {
		common::Streams::err << "Unknown exception caught." << std::endl;
		return 127;
	}
}
